#!/usr/bin/perl -w
use strict;
use warnings;

use Carp;
use Data::Dumper;
use Try::Tiny;

#####START MODIFY: SET ALL PATHS CORRECTLY BELOW

use GetSMotifsfromPDB qw(missing_residues extract_loops);
# Directory that holds the pdb files to be parsed
# If entire PDB is in one directory, use the same path for both
# $PDB_DIR and $PDB_OBSOLETES. 
my $PDB_DIR       = "/usr/local/databases/pdb/uncompressed/";
my $PDB_OBSOLETES = "/usr/local/databases/pdb/obsoletes/";
$GetSMotifsfromPDB::PDB_DIR       = $PDB_DIR;
$GetSMotifsfromPDB::PDB_OBSOLETES = $PDB_OBSOLETES;

#PATH for DSSP
my $DSSP_PATH = "/usr/local/bin";
my $DSSP_EXEC = "dsspcmbi";
$GetSMotifsfromPDB::DSSP_PATH = $DSSP_PATH;
$GetSMotifsfromPDB::DSSP_EXEC = $DSSP_EXEC;

use FixLoopsWithModeller qw(remodel 
		renumber_pdb
        get_CA_only_chains
        get_gap_list
         );
#Directories where output file will be stored
my $MODELLER_LOG_FILES = "/tmp/modeller/modeller_log_files";
my $MODELLER_OUT_FILES = "/tmp/modeller/remodeled_pdb/";
my $MODELLER_PIR_FILES = "/tmp/modeller/modeller_pir_files";
$FixLoopsWithModeller::MODELLER_LOG_FILES = $MODELLER_LOG_FILES;
$FixLoopsWithModeller::MODELLER_OUT_FILES = $MODELLER_OUT_FILES;
$FixLoopsWithModeller::MODELLER_PIR_FILES = $MODELLER_PIR_FILES;

#PATH for Modeller
my $MODELLER_PATH = "/usr/bin";
my $MODELLER_EXEC = "mod9.14";
$FixLoopsWithModeller::MODELLER_PATH = $MODELLER_PATH;
$FixLoopsWithModeller::MODELLER_EXEC = $MODELLER_EXEC;

#PATH for EMBOSS NEEDLE
my $EMBOSS_PATH = "/usr/local/bin";
my $NEEDLE_EXEC = "needle";
$FixLoopsWithModeller::EMBOSS_PATH = $EMBOSS_PATH;
$FixLoopsWithModeller::NEEDLE_EXEC = $NEEDLE_EXEC;

###END ALL PATHS SET

####WARNING: DO NOT MODIFY ANYTHING BEYOND THIS POINT

my %WARNS;
local $SIG{__WARN__} = sub {
    my $message = shift;
    return if $WARNS{$message}++; # to reduce the number of copies of the same warning 
    logger('warning', $message);
};
local $SIG{__DIE__} = sub {
    my $message = shift;
    logger('error', $message);
};

#  PDB code and chain input
my $pdb_code  = $ARGV[0];
my $chain     = $ARGV[1];

die "pdb_code is required\n" unless $pdb_code;
die "chain is required\n"    unless defined $chain;

#output files
my $outputlogfile = $pdb_code . "_" . $chain . ".remodel_pdb.log";

#open outputs
#open LOG, ">$outputlogfile" || die "can't open $outputlogfile log file\n";

#
# get the smotifs
# check for missing residue in the original pdb file
#eval {
#START: try {
try {
	my $new_pdb_file;
	
		#print "$pdb_code\t$chain\t$PDB_DIR\t$PDB_OBSOLETES\n";        

        my @ca_chains = get_CA_only_chains(        
		pdb                => $pdb_code,
		chain              => $chain,
		pdb_path           => $PDB_DIR,
		pdb_obsoletes_path => $PDB_OBSOLETES
        );
        # print Dumper(\@ca_chains) if @ca_chains; 
        # if pdb contains just C-alpha
        # then remodel the whole structure
        # if (@ca_chains &&  exists_on_array($chain, \@ca_chains) ) {
        if ( exists_on_array($chain, \@ca_chains) ) {
                print "CHECKING FOR SEQRES GAPS IN PDB WITH CA ONLY CHAIN $pdb_code $chain\n";
		my ($mismatch, $gap_in_seqres, $aref_new_gap_list) = get_gap_list(
				pdb                => $pdb_code,
				chain              => $chain,
				pdb_path           => $PDB_DIR,
				pdb_obsoletes_path => $PDB_OBSOLETES
			    );
	      
		if ($gap_in_seqres) {
		    die "$pdb_code $chain: Alignment error: Gaps in SEQRES";
		}              
		if ($mismatch) {
		    die "$pdb_code $chain: Alignment error: alignment mismatch";
		}              
            print "REMODELING WHOLE PDB WITH CA ONLY CHAIN $pdb_code $chain\n";
	    $new_pdb_file = remodel (
		pdb                => $pdb_code,
		chain              => $chain,
		pdb_path           => $PDB_DIR,
		pdb_obsoletes_path => $PDB_OBSOLETES,
                structure          => 'whole'
	    );
        } else {  
		# Look for missing residues (using Joe's approach) and disordered proteins
		# if proteins is disordered then missing == 0
		my ($missing, $disordered) = missing_residues( $pdb_code, $chain );
                if ($disordered) {
		    
                    print "RENUMBERING A DISORDERED PDB $pdb_code $chain\n";
		    $new_pdb_file = renumber_pdb(
			pdb                => $pdb_code,
			chain              => $chain,
			pdb_path           => $PDB_DIR,
			pdb_obsoletes_path => $PDB_OBSOLETES
		    );
                
                } else {
			# look for missing residues(gaps list) and alignment errors

			my ($mismatch, $gap_in_seqres, $aref_new_gap_list) = get_gap_list(
					pdb                => $pdb_code,
					chain              => $chain,
					pdb_path           => $PDB_DIR,
					pdb_obsoletes_path => $PDB_OBSOLETES
				    );
                      
			if ($gap_in_seqres) {
			    die "$pdb_code $chain: Alignment error: Gaps in SEQRES";
			}              
			if ($mismatch) {
			    die "$pdb_code $chain: Alignment error: alignment mismatch";
			}              
			
			if (@$aref_new_gap_list) { # If there are missing residues, remodel loops
			
			    print "REMODELING _PDB_WITH_MISSING_RESIDUES $pdb_code $chain\n";
			    # remodel_pdb_with_missing_residues;
			    $new_pdb_file = remodel (
				pdb                => $pdb_code,
				chain              => $chain,
				pdb_path           => $PDB_DIR,
				pdb_obsoletes_path => $PDB_OBSOLETES,
				structure          => 'loop'
			    );
			    # print new pdb filename or location
			}
			else { # No missing residues, hence just renumber
			    
			    print "RENUMBERING PDB $pdb_code $chain\n";
			    $new_pdb_file = renumber_pdb(
				pdb                => $pdb_code,
				chain              => $chain,
				pdb_path           => $PDB_DIR,
				pdb_obsoletes_path => $PDB_OBSOLETES
			    );

			}
                 }
        }

	# print new pdb filename or location
	print "NEW_PDB_FILE=$new_pdb_file\n";
	print "\n\nDONE!!\n\n";

	unlink "$pdb_code".'_'."$chain.remodel_pdb.log";
	unlink "$pdb_code".$chain.".dssp.out";
	unlink "$pdb_code".$chain.".dssp";
	unlink "$pdb_code".$chain.".pdb"; 
	unlink "q";
	unlink glob "*.log";
       
}
catch {
    #print "There was error $@";
    print "There was error: $_\n";
}; 
#close(LOG) or die "can't close log file\n";
#close(OUT) or die "can't close output file\n";

exit;

sub exists_on_array {
    my ($string, $aref) = @_;
 
    return 0 unless @$aref;
    return 0 unless $string;
   
    chomp $string;
    
    $string =~ s/^\s+//g;
    $string =~ s/\s+$//g;
    foreach my $element (@$aref) {
        $element =~ s/^\s+//g;
        $element =~ s/\s+$//g;
        return 1 if $element eq $string;
    }
    return 0; 
}

sub logger {
	my ($level, $msg) = @_;
	if (open my $out, '>>', '/tmp/log.txt') {
		flock ($out, 2 ); ## exclusive lock
                chomp $msg;
		print $out "$level - $msg\n";
                close $out;
        }
}

=head1 NAME

ALGORITHM TO REMODEL MISSING RESIDUES IN PDB STRUCTURES

=head1 SYNOPSIS

This program will find and REMODEL missing residues in a protein 
structure using MODELLER. If no missing residues are found in
the structure, it will simply renumber the structure to fix 
insertion codes. CA-only chains will be remodeled completely. 
Disordered proteins will just be renumbered.

=head1 PRE-REQUISITES

The following sofware / Data are required: 

1. Modeller (version 9.14 https://salilab.org/modeller/)

2. DSSP (http://swift.cmbi.ru.nl/gv/dssp/)

3. EMBOSS (http://emboss.sourceforge.net/)

4. Local PDB directory (central or user-designated from http://www.rcsb.org).

Download and install the above mentioned software / data according to their instructions.

Non-standard perl modules required:

1. Exporter

2. Carp

3. Data::Dumper

4. Try::Tiny

5. File::Copy

6. File::Path

7. File::Basename

8. File::Find::Rule

These perl modules can be downloaded and installed from CPAN (http://www.cpan.org/) using a CPAN client.

=head1 IMPORTANT NOTE

Modify lines 9 to 53 in this script to fix all the paths correctly. 

(For Modeller, DSSP, EMBOSS, PDB, OUTPUT_DIRECTORIES)

=head1 USAGE

perl remodel_pdb.pl <pdb_code> <chain_id>

Example: perl remodel_pdb.pl 1abc A

=head1 OUTPUT

pdb1abcA.ent will be generated and stored in the output directory 
specified in the path (to be set in lines 9 to 53 in remodel_pdb.pl). 
Modeller log files and pir files will also be stored in the directories 
specified. These can be used for error-checking and can be removed later.
Note that all files have the chain names because remodeling is done
chain-wise. 

=head1 OPTIONS

=over 8

=item B<-help>

Print a brief help message and exits.

=item B<-man>

Prints the manual page and exits.

=back

=head1 AUTHORS

Brinda Vallat, Joseph Dybas, Carlos Madrid and Andras Fiser.

=head1 DESCRIPTION

This program will find and REMODEL missing residues in a protein 
structure using MODELLER. If no missing residues are found in
the structure, it will simply renumber the structure to fix 
insertion codes. CA-only chains will be remodeled completely. 
Disordered proteins will just be renumbered. 

=cut
